#include <SDL.h>
#include <string.h>
#include <stdlib.h>

#include "cube.h"

#define MAXCONLINE 100
#define MAXLINES 50

char conbuf[MAXLINES][MAXCONLINE];
char *conrefs[MAXLINES];
int conused = 0;
int ndraw = 5;

void conout(char *s, int len)
{
    if(!conused) for(int i = 0; i<MAXLINES; i++) conrefs[i] = conbuf[i];
    char *buf = conrefs[MAXLINES-1];
    for(int i = MAXLINES-1; i>0; i--) conrefs[i] = conrefs[i-1];
    conrefs[0] = buf;
    int l = min(len, MAXCONLINE-1);
    strncpy(buf, s, l);
    buf[l] = 0;
    if(conused<MAXLINES) conused++;
};

void conoutf(char *s, int a, int b, int c)
{
    char buf[1024];
    sprintf(buf, s, a, b, c);
    conout(buf, strlen(buf)); 
};

void renderconsole()
{
    int nd = min(conused, ndraw);
    for(int i = 0; i<nd; i++)
    {
        draw_text(conrefs[i], 10, 25*(nd-i-1)+10, 2, 0, 0);
    };
};

struct keym { int code; char *name; char *action; } keyms[] =
{                                                                
    SDLK_BACKSPACE  , "BACKSPACE",     "",       
    SDLK_TAB        , "TAB",           "",       
    SDLK_CLEAR      , "CLEAR",         "",       
    SDLK_RETURN     , "RETURN",        "",       
    SDLK_PAUSE      , "PAUSE",         "",       
    SDLK_ESCAPE     , "ESCAPE",        "quit",       
    SDLK_SPACE      , "SPACE",         "",       
    SDLK_EXCLAIM    , "EXCLAIM",       "",       
    SDLK_QUOTEDBL   , "QUOTEDBL",      "",       
    SDLK_HASH       , "HASH",          "",       
    SDLK_DOLLAR     , "DOLLAR",        "",       
    SDLK_AMPERSAND  , "AMPERSAND",     "",       
    SDLK_QUOTE      , "QUOTE",         "",       
    SDLK_LEFTPAREN  , "LEFTPAREN",     "",       
    SDLK_RIGHTPAREN , "RIGHTPAREN",    "",       
    SDLK_ASTERISK   , "ASTERISK",      "",       
    SDLK_PLUS       , "PLUS",          "",       
    SDLK_COMMA      , "COMMA",         "equalize 0",       
    SDLK_MINUS      , "MINUS",         "",       
    SDLK_PERIOD     , "PERIOD",        "equalize 2",       
    SDLK_SLASH      , "SLASH",         "",       
    SDLK_0          , "0",             "",       
    SDLK_1          , "1",             "",       
    SDLK_2          , "2",             "",       
    SDLK_3          , "3",             "",       
    SDLK_4          , "4",             "",       
    SDLK_5          , "5",             "",       
    SDLK_6          , "6",             "",       
    SDLK_7          , "7",             "",       
    SDLK_8          , "8",             "",       
    SDLK_9          , "9",             "",       
    SDLK_COLON      , "COLON",         "",       
    SDLK_SEMICOLON  , "SEMICOLON",     "",       
    SDLK_LESS       , "LESS",          "",       
    SDLK_EQUALS     , "EQUALS",        "",       
    SDLK_GREATER    , "GREATER",       "",       
    SDLK_QUESTION   , "QUESTION",      "",       
    SDLK_AT         , "AT",            "",       
    SDLK_LEFTBRACKET, "LEFTBRACKET",   "editheight 0 1",       
    SDLK_BACKSLASH  , "BACKSLASH",     "",       
    SDLK_RIGHTBRACKET,"RIGHTBRACKET",  "editheight 0 -1",       
    SDLK_CARET      , "CARET",         "",       
    SDLK_UNDERSCORE , "UNDERSCORE",    "",       
    SDLK_BACKQUOTE  , "BACKQUOTE",     "",       
    SDLK_a          , "a",             "",       
    SDLK_b          , "b",             "",       
    SDLK_c          , "c",             "solid 0",       
    SDLK_d          , "d",             "",       
    SDLK_e          , "e",             "edittoggle",       
    SDLK_f          , "f",             "solid 1",       
    SDLK_g          , "g",             "",       
    SDLK_h          , "h",             "say hi",       
    SDLK_i          , "i",             "",       
    SDLK_j          , "j",             "say blah; say :)",       
    SDLK_k          , "k",             "",       
    SDLK_l          , "l",             "newentity light 16",       
    SDLK_m          , "m",             "",       
    SDLK_n          , "n",             "",       
    SDLK_o          , "o",             "editheight 2 1",       
    SDLK_p          , "p",             "editheight 2 -1",       
    SDLK_q          , "q",             "",       
    SDLK_r          , "r",             "recalc",       
    SDLK_s          , "s",             "",       
    SDLK_t          , "t",             "",       
    SDLK_u          , "u",             "undo",       
    SDLK_v          , "v",             "",       
    SDLK_w          , "w",             "",       
    SDLK_x          , "x",             "clearents light",       
    SDLK_y          , "y",             "",       
    SDLK_z          , "z",             "",       
    SDLK_DELETE     , "DELETE",        "edittex 0 -1",       
    SDLK_KP0        , "KP0",           "",       
    SDLK_KP1        , "KP1",           "",       
    SDLK_KP2        , "KP2",           "",       
    SDLK_KP3        , "KP3",           "",       
    SDLK_KP4        , "KP4",           "",       
    SDLK_KP5        , "KP5",           "",       
    SDLK_KP6        , "KP6",           "",       
    SDLK_KP7        , "KP7",           "",       
    SDLK_KP8        , "KP8",           "",       
    SDLK_KP9        , "KP9",           "",       
    SDLK_KP_PERIOD  , "KP_PERIOD",     "",       
    SDLK_KP_DIVIDE  , "KP_DIVIDE",     "",       
    SDLK_KP_MULTIPLY, "KP_MULTIPLY",   "",       
    SDLK_KP_MINUS   , "KP_MINUS",      "",       
    SDLK_KP_PLUS    , "KP_PLUS",       "",       
    SDLK_KP_ENTER   , "KP_ENTER",      "",       
    SDLK_KP_EQUALS  , "KP_EQUALS",     "",       
    SDLK_UP         , "UP",            "forward",       
    SDLK_DOWN       , "DOWN",          "backward",       
    SDLK_RIGHT      , "RIGHT",         "right",       
    SDLK_LEFT       , "LEFT",          "left",       
    SDLK_INSERT     , "INSERT",        "edittex 0 1",       
    SDLK_HOME       , "HOME",          "edittex 1 1",       
    SDLK_END        , "END",           "edittex 1 -1",       
    SDLK_PAGEUP     , "PAGEUP",        "edittex 2 1",       
    SDLK_PAGEDOWN   , "PAGEDOWN",      "edittex 2 -1",       
    SDLK_F1         , "F1",            "locklod 2",       
    SDLK_F2         , "F2",            "locklod 10",       
    SDLK_F3         , "F3",            "locklod 20",       
    SDLK_F4         , "F4",            "locklod 50",       
    SDLK_F5         , "F5",            "fpsrange 64 80",       
    SDLK_F6         , "F6",            "fpsrange 32 40",       
    SDLK_F7         , "F7",            "fpsrange 16 20",       
    SDLK_F8         , "F8",            "fpsrange 8 10",       
    SDLK_F9         , "F9",            "fpsrange 4 5",       
    SDLK_F10        , "F10",           "",       
    SDLK_F11        , "F11",           "",       
    SDLK_F12        , "F12",           "screenshot",       
    SDLK_F13        , "F13",           "",       
    SDLK_F14        , "F14",           "",       
    SDLK_F15        , "F15",           "",       
    SDLK_NUMLOCK    , "NUMLOCK",       "",       
    SDLK_CAPSLOCK   , "CAPSLOCK",      "",       
    SDLK_SCROLLOCK  , "SCROLLOCK",     "",       
    SDLK_RSHIFT     , "RSHIFT",        "",       
    SDLK_LSHIFT     , "LSHIFT",        "",       
    SDLK_RCTRL      , "RCTRL",         "",       
    SDLK_LCTRL      , "LCTRL",         "",       
    SDLK_RALT       , "RALT",          "",       
    SDLK_LALT       , "LALT",          "",       
    SDLK_RMETA      , "RMETA",         "",       
    SDLK_LMETA      , "LMETA",         "",       
    SDLK_LSUPER     , "LSUPER",        "",       
    SDLK_RSUPER     , "RSUPER",        "",       
    SDLK_MODE       , "MODE",          "",       
    SDLK_COMPOSE    , "COMPOSE",       "",       
    SDLK_HELP       , "HELP",          "",       
    SDLK_PRINT      , "PRINT",         "",       
    SDLK_SYSREQ     , "SYSREQ",        "",       
    SDLK_BREAK      , "BREAK",         "",       
    SDLK_MENU       , "MENU",          "",       
    SDLK_MENU       , "MENU",          "", 
    0      
};                                                               

void bind(char *key, int klen, char *action, int alen)
{
    key[klen] = 0;  // fucks up recursive bind
    action[alen] = 0;
    keym *p;
    for(p = keyms; p->code; p++) if(strcmp(p->name, key)==0) break;
    if(!p->code)
    {
        conoutf("unknown key \"%s\"", (int)key);   
        return;
    };
    p->action = action;
};

int findtype(char *what, int len)
{
    int c = what[len];
    what[len] = 0;
    int t = NOTUSED;
    if(strcmp(what,"light")==0) t = LIGHT;
    else if(strcmp(what,"playerstart")==0) t = PLAYERSTART;
    else conoutf("unknow entity type \"%s\"", (int)what);
    what[len] = c;
    return t;
}

void delent(char *what, int len)
{
    switch(findtype(what, len))
    {
        case LIGHT: clearents(LIGHT); break;
    };
};

void newent(char *what, int len, int val)
{
    switch(findtype(what, len))
    {
        case LIGHT:
        {
            entity *e = newentity((int)player1->x, (int)player1->z, (int)player1->y, LIGHT);
            e->attr = val;
        };
    };
};

void parseword(char *&p, char *&word, int &len)
{
    p += strspn(p, " \t\r");
    if(*p=='\"')
    {
        p++;
        word = p;
        len = strcspn(p, "\"\r\n\0");
        if(*p=='\"') p++;
    }
    else
    {
        word = p;
        len = strcspn(p, "; \t\r\n\0");
    };
    p += len;
};

void unknown(char *c, bool isdown)
{
    if(isdown) conoutf("unknown command \"%s\"", (int)c);
};

void execute(char *p, bool down = true)
{
    for(bool cont = true; cont;)
    {
        char *command, *a1, *a2, *a3, c[256];
        int clen, a1len, a2len, a3len;

        parseword(p, command, clen);
        parseword(p, a1, a1len);
        parseword(p, a2, a2len);
        parseword(p, a3, a3len);

        p += strcspn(p, ";\n\0");
        cont = *p++!=0;
        if(!clen) continue;

        bool isdown = down;
        if(command[0]=='+') { isdown = true; command++; };
        if(command[0]=='-') { isdown = false; command++; };
        if(command[0]=='/') command++; // irc style
        if(command[0]=='/') continue;  // it's a comment

        strncpy(c, command, clen);
        c[clen] = 0;

        switch(c[0])
        {
            #define s(a) case a: if(false) void (0);
            #define both(n,f) else if(strcmp(c, n)==0) (f)
            #define once(n,f) else if(strcmp(c, n)==0) (isdown ? (f) : void (0))
            #define last() else unknown(c, isdown); break
            #define dir(v,d) v = isdown ? d : (v==-(d) ? -(d) : 0)

            s('b') both("backward", dir(move, -1));
                   once("bind",     bind(a1, a1len, a2, a2len));
                   last();

            s('c') once("clearents",  delent(a1,a1len));
                   last();

            s('e') once("edittex",    edittex(atoi(a1), atoi(a2)));
                   once("edittoggle", (editmode = !editmode));
                   once("editheight", editheight(atoi(a1)==0, atoi(a2)));
                   once("equalize",   editequalise(atoi(a1)==0));
                   last();

            s('f') once("fpsrange",   fpsrange(atof(a1), atof(a2)));
                   both("forward",    dir(move, 1));
                   last();

            s('l') both("left",       dir(strafe, 1));
                   both("locklod",    locklod(atoi(a1)*(int)isdown));
                   last();

            s('n') once("newentity",  newent(a1,a1len,atoi(a2)));
                   last();

            s('q') once("quit",       quit());
                   last();

            s('r') both("right",      dir(strafe, -1));
                   once("recalc",     calclight());
                   last();

            s('s') once("screenshot", screenshot());
                   once("solid",      editfill(atoi(a1)!=0));
                   once("say",        conout(a1, a1len));
                   last();

            s('u') once("undo",       editundo());
                   last();

            default:
                unknown(c, isdown);
        };
    };
};

void keypress(int code, bool isdown)
{
    for(keym *p = keyms; p->code; p++) if(p->code==code)
    {
        execute(p->action, isdown);
        return;
    };
};

void writecfg()
{
    FILE *f = fopen("config.cfg", "w");
    if(!f) return;
    fprintf(f, "// generated automatically, will be read & overwritten\n\n");
    for(keym *p = keyms; p->code; p++) fprintf(f, "bind %s \"%s\"\n", p->name, p->action);
    fclose(f);
};

char *loadfile(char *fn)
{
    FILE *f = fopen(fn, "rb");
    if(!f) return NULL;
    fseek(f, 0, SEEK_END);
    int len = ftell(f);
    fseek(f, 0, SEEK_SET);
    char *buf = (char *)malloc(len+1);
    if(!buf) return NULL;
    buf[len] = 0;
    int rlen = fread(buf, 1, len, f);
    fclose(f);
    if(len!=rlen || len<=0) return NULL;
    return buf;
};

void exec(char *cfgfile)
{
    char *buf = loadfile(cfgfile); // never dealloc
    if(!buf)
    {
        conoutf("could not read \"%s\"", (int)cfgfile);
        return;
    };
    execute(buf);
};
