typedef unsigned char uchar;

enum                // block types
{
    SOLID = 0,      // entirely solid cube [only specifies wtex]
    CORNER,         // half full corner of a wall
    SLOPE,          // half full slope on floor
    SUPPORT,        // half full support on ceiling
    SPACE           // entirely empty cube
};

struct sqr
{
    uchar type;
    char floor, ceil;
    uchar wtex, ftex, ctex;
    uchar light;
    //uchar special;
};

enum                // entity types
{
    NOTUSED = 0,    // entity slot not in use in map
    LIGHT,          // lightsource, attr = intensity
    PLAYERSTART     // attr = angle
};

struct entity       // static ent
{
    short x, y, z;
    short attr;
    uchar type, pad1, pad2, pad3;
};

#define MAXENTS 1024
#define MAPVERSION 1

struct header
{
    char head[4];               // "CUBE"
    int version;                // any >8bit quantity is a little indian
    int headersize;             // sizeof(header)
    int sfactor;                // in bits
    char maptitle[128];
    uchar texlists[3][256];
    entity ents[MAXENTS];
    int extension[256];         // for adding stuff without disturbing the format
    char tail[4];               // "EBUC"
    // map array follows here, size = ((1<<sfactor)^2)*sizeof(sqr)
};

#define SWS(w,x,y,s) (&(w)[(y)*(s)+(x)])
#define S(x,y) SWS(world,x,y,size)
#define MIP(n) (&world[csize*2-((csize*2)>>(n))])
#define SMALLEST_FACTOR 6       // determines number of mips there can be
#define DEFAULT_FACTOR 8
#define LARGEST_FACTOR 12       // 10 is already insane
#define SOLID(x) ((x)->type==SOLID)
#define MINBORD 2
#define OUTBORD(x,y) ((x)<MINBORD || (y)<MINBORD || (x)>=size-MINBORD || (y)>=size-MINBORD)

struct dynent                   // players, monsters, projectiles etc
{
    float x, y, z;
    float yaw, pitch, roll;
    float maxspeed;             // cubes per second
    bool outsidemap;            // from his eyes
};

// globals ooh

extern sqr *world;
extern header *hdr;
extern int sfactor, size, csize;
extern dynent *player1;         // client ent that receives input and acts as camera
extern bool editmode;
extern int move, strafe;


#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))

#ifdef WIN32
#define PATHDIV '\\'
#else
#define PATHDIV '/'
#endif

extern void quit();
extern void pre_gfx_init();
extern void gl_init(int w, int h);
extern void cleangl();
extern void gl_drawframe(int w, int h, int changelod, float curfps);
extern void mousemove(int dx, int dy); 
extern void mouseclick(int isdown, int butnum); 
extern void updateworld(int millis);
extern void empty_world();
extern void remip(int level, int x, int y, int xs, int ys);
extern void render_world(int vx, int vy, int vh, int changelod, int yaw);
extern void render_flat(int tex, int x, int y, int size, int h, uchar l1, uchar l2, uchar l3, uchar l4);
extern void render_square(int wtex, int floor, int ceil, int x1, int y1, int x2, int y2, int size, uchar l1, uchar l2);
extern void fatal(char *s, char *o = "");
extern void *alloc(int s);
extern int lod_factor();
extern void draw_text(char *fstr, int left, int top, int gl_num, int arg1, int arg2);
extern void installtex(int tnum, char *texname, int &xs, int &ys);
extern void cursorupdate(double x, double y, double z);
extern void line(int x1, int y1, int z1, int x2, int y2, int z2);
extern void box(int x, int y, int z, int w, int h);
extern void linestyle(float width, int r, int g, int b);
extern void editdrag(bool isdown);
extern void editheight(bool isfloor, int amount);
extern void edittex(int type, int dir);
extern void editfill(bool solid);
extern void editequalise(bool isfloor);
extern void editundo();
extern void keypress(int code, bool isdown);
extern void fpsrange(double low, double high);
extern void screenshot();
extern void writecfg();
extern void renderconsole();
extern void conoutf(char *s, int a = 0, int b = 0, int c = 0);
extern void exec(char *cfgfile);
extern void locklod(int f);
extern entity *newentity(int x, int y, int z, int type);
extern void clearents(int type);
extern void calclight();

