//#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cube.h"

int cx, cy, ch;
#define GRIDSIZE 5
#define GRIDW 0.5f
#define GRID8 2.0f
#define GRIDS 2.0f
#define GRIDM 0x7
int selx, sely, selh, selxs, selys;
bool selset = false;
int curedittex[] = { -1, -1, -1 };

void cursorupdate(double x, double y, double z)
{
    cx = (int)x;
    cy = (int)y;

    if(OUTBORD(cx, cy)) return;

    sqr *s = S(cx,cy);

    for(int ix = cx-GRIDSIZE; ix<=cx+GRIDSIZE; ix++) for(int iy = cy-GRIDSIZE; iy<=cy+GRIDSIZE; iy++)
    {
        if(OUTBORD(ix, iy)) continue;
        sqr *s = S(ix,iy);
        if(SOLID(s)) continue;
        int ih = z-s->floor<s->ceil-z ? s->floor : s->ceil;
        linestyle(GRIDW, 0x80, 0x80, 0x80);
        box(ix, iy, ih, 1, 1);
        linestyle(GRID8, 0x40, 0x40, 0xFF);
        if(!(ix&GRIDM))   line(ix,   iy,   ih, ix,   iy+1, ih);
        if(!(ix+1&GRIDM)) line(ix+1, iy,   ih, ix+1, iy+1, ih);
        if(!(iy&GRIDM))   line(ix,   iy,   ih, ix+1, iy,   ih);
        if(!(iy+1&GRIDM)) line(ix,   iy+1, ih, ix+1, iy+1, ih);
    };

    if(!SOLID(s))
    {
        ch = z-s->floor<s->ceil-z ? s->floor : s->ceil;
        linestyle(GRIDS, 0xFF, 0xFF, 0xFF);
        box(cx, cy, ch, 1, 1);
    };

    if(selset)
    {
        linestyle(GRIDS, 0xFF, 0x40, 0x40);
        box(selx, sely, selh, selxs, selys);
    };
};

void postedit()
{
    remip(0, selx, sely, selxs, selys);
};

#define MAXUNDOSIZE 64
sqr undobuf[MAXUNDOSIZE][MAXUNDOSIZE];
int ux, uy, uys, us;
bool uset = false;

void preedit()
{
    if(selxs>MAXUNDOSIZE || selys>MAXUNDOSIZE) return;
    sqr *s = S(selx, sely);
    us = sizeof(sqr)*selxs;
    for(int y = 0; y<selys; y++)
    {
        memcpy(undobuf[y], &s[y*size], us);
    };
    uys = selys;
    ux = selx;
    uy = sely;
    uset = true;
};

void editundo()
{
    if(!uset) { conoutf("nothing more to undo"); return; };
    sqr *s = S(ux, uy);
    for(int y = 0; y<uys; y++)
    {
        memcpy(&s[y*size], undobuf[y], us);
    };
    uset = false;
    postedit();
};

void deselect()
{
    selset = false;
    for(int i = 0; i<3; i++)
    {
        int c = curedittex[i];
        if(c>=0)
        {
            uchar *p = hdr->texlists[i];
            int t = p[c];
            for(int a = c-1; a>=0; a--) p[a+1] = p[a];
            p[0] = t;
            curedittex[i] = -1;
        };
    };
};

void editdrag(bool isdown)
{
    static int lastx, lasty, lasth;
    if(isdown)
    {
        lastx = cx;
        lasty = cy;
        lasth = ch;
        deselect();
    }
    else
    {
        selx = min(lastx,cx);
        sely = min(lasty,cy);
        selh = max(lasth,ch);
        selxs = abs(lastx-cx)+1;
        selys = abs(lasty-cy)+1;
        selset = true;
        if(OUTBORD(selx, sely)) selset = false;
        int bsize = size-MINBORD;
        if(selxs+selx>bsize) selxs = bsize-selx;
        if(selys+sely>bsize) selys = bsize-sely;
    };
};

void editheight(bool isfloor, int amount)
{
    if(!selset) return;
    preedit();
    for(int x = 0; x<selxs; x++) for(int y = 0; y<selys; y++)
    {
        sqr *s = S(selx+x, sely+y);
        if(isfloor)
        {
            s->floor += amount;
            if(s->floor>=s->ceil) s->floor = s->ceil-1;
        }
        else
        {
            s->ceil += amount;
            if(s->ceil<s->floor) s->ceil = s->floor+1;
        };
    };
    postedit();
};

void edittex(int type, int dir)
{
    if(!selset || type<0 || type>2) return;
    preedit();
    int i = curedittex[type];
    i = i<0 ? 0 : i+dir;
    curedittex[type] = i = min(max(i, 0), 255);
    int t = hdr->texlists[type][i];

    for(int x = 0; x<selxs; x++) for(int y = 0; y<selys; y++)
    {
        sqr *s = S(selx+x, sely+y);
        switch(type)
        {
            case 0: s->ftex = t; break;
            case 1: s->wtex = t; break;
            case 2: s->ctex = t; break;
        };
    };
    postedit();
};

void editfill(bool solid)
{
    if(!selset) return;
    preedit();
    for(int x = 0; x<selxs; x++) for(int y = 0; y<selys; y++)
    {
        S(selx+x, sely+y)->type = solid ? SOLID : SPACE;
    };
    postedit();
};

void editequalise(bool isfloor)
{
    if(!selset) return;
    preedit();
    int low = 127, hi = -128;

    for(int a = 0; a<selxs; a++) for(int b = 0; b<selys; b++)
    {
        sqr *s = S(selx+a, sely+b);
        if(s->floor<low) low = s->floor;
        if(s->ceil>hi) hi = s->ceil;
    };

    for(int x = 0; x<selxs; x++) for(int y = 0; y<selys; y++)
    {
        sqr *s = S(selx+x, sely+y);
        if(isfloor) s->floor = low; else s->ceil = hi;
        if(s->floor>=s->ceil) s->floor = s->ceil-1;
    };

    postedit();
};

