#ifdef WIN32
#include "windows.h"
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include "SDL.h"

#include "cube.h"

#include <stdlib.h>

double minfps = 32.0;
double maxfps = 40.0;

int lastticks = 0;
int frames = 0;
double fps = 0.0;

int scr_w = 512; int scr_h = 384;
//int scr_w = 320; int scr_h = 240;
//int scr_w = 800; int scr_h = 600;

void cleanup()                  // single program exit point;
{
    cleangl();
    SDL_ShowCursor(1);
    SDL_Quit();
    exit(1);
};

void quit()                     // normal exit
{
    writecfg();
    cleanup();
};

void fatal(char *s, char *o)    // failure exit
{
    printf("%s%s (%s)\n", s, o, SDL_GetError());
    cleanup();
};

void *alloc(int s)
{
    void *b = calloc(1,s);
    if(!b) fatal("out of memory!");
    return b;
};

void screenshot()
{
    SDL_Surface *image;
    SDL_Surface *temp;
    int idx;
    if(image = SDL_CreateRGBSurface(SDL_SWSURFACE, scr_w, scr_h, 24, 0x0000FF, 0x00FF00, 0xFF0000, 0))
    {
        if(temp  = SDL_CreateRGBSurface(SDL_SWSURFACE, scr_w, scr_h, 24, 0x0000FF, 0x00FF00, 0xFF0000, 0))
        {
            glReadPixels(0, 0, scr_w, scr_h, GL_RGB, GL_UNSIGNED_BYTE, image->pixels);
            for (idx = 0; idx<scr_h; idx++)
            {
                memcpy((char *)temp->pixels+3*scr_w*idx, (char *)image->pixels+3*scr_w*(scr_h-1-idx), 3*scr_w);
            };
            SDL_SaveBMP(temp, "screenshot.bmp");
            SDL_FreeSurface(temp);
        };
        SDL_FreeSurface(image);
    };
};

void calcfps()
{
    frames++;
    int ticks = SDL_GetTicks();
    fps = 1000.0/((ticks-lastticks));
    lastticks = ticks;
};

void fpsrange(double low, double high)
{
    if(low>high || low<1.0 || high>200.0) return;
    minfps = low;
    maxfps = high;
};

char *path(char *s)
{
    for(char *t = s; t = strchr(t, '/'); *t = PATHDIV);
    return s;
};

int main(int argc, char **argv)
{  
    pre_gfx_init();

    if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER)<0) fatal("Unable to initialize SDL");
    if(SDL_SetVideoMode(scr_w, scr_h, 0, SDL_OPENGL|SDL_FULLSCREEN)==NULL) fatal("Unable to create OpenGL screen");

    SDL_WM_SetCaption("cube engine", NULL);
    SDL_ShowCursor(0);

    gl_init(scr_w, scr_h);

    int xs, ys;
    installtex(2, path("data/don_chars.bmp"), xs, ys);
    installtex(1, path("data/crosshair.bmp"), xs, ys);

    int ignore = 5;

    updateworld(SDL_GetTicks());

    exec("config.cfg");
    exec("autoexec.cfg");

    for(;;)
    {
        updateworld(SDL_GetTicks());
        gl_drawframe(scr_w, scr_h, fps<minfps ? -1 : (fps>maxfps ? 1 : 0), (float)fps);
        calcfps();
        SDL_GL_SwapBuffers();

        SDL_Event event;
        while ( SDL_PollEvent(&event) )
        {
            switch( event.type )
            {
                case SDL_QUIT:
                    quit();
                    break;

                case SDL_KEYDOWN: 
                case SDL_KEYUP: 
                    keypress(event.key.keysym.sym, event.key.state==SDL_PRESSED);
                    break;

                case SDL_MOUSEMOTION:
                    if(ignore) { ignore--; break; };
                    mousemove(event.motion.xrel, event.motion.yrel);
                    break;

                case SDL_MOUSEBUTTONDOWN:
                case SDL_MOUSEBUTTONUP:
                    mouseclick(event.button.state, event.button.button);
                    break;
            };
        };
    };
    quit();
    return 1;
};
