// the engine

#include "cube.h"

#ifdef WIN32
#include "windows.h"
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>
#include <glh/glh_extensions.h>


#include <stdio.h>
#include <math.h>

#include <SDL.h>

void *alloc(int s)
{
    void *b = calloc(1,s);
    if(!b) fatal("out of memory!");
    return b;
};

double camx = 128.0, camy = 5.0, camz = 128.0;
double mouseyaw = 270.0, mousepitch = 0.0;
double mousesens = 10.0;
bool fisheye = false;

void pre_gfx_init()
{
    empty_world();
    camx = size/2;
    camz = size/2;
};

int move = 0, strafe = 0;
bool editmode = false; 

class player_entity
{
    public:
    double x, y;
    double h;
    double yaw, pitch, roll;
    double maxspeed;

    player_entity()
    {
        x = 0; y = 0; h = 1; yaw = 0; pitch = 0; roll = 0; maxspeed = 0.1;
    };

    void move(double forward, double sideways, int time)
    {
        double xf = cos(yaw);
        double yf = sin(yaw);
        double nx = xf*forward*time+x;  // and strafe
        double ny = yf*forward*time+y;
    };

    void input(int m, int strafe, int time, double nyaw, double npitch) // take input
    {
        yaw = nyaw; pitch = npitch;
        move(m*maxspeed, strafe*maxspeed, time);
    };
};

player_entity *player1 = new player_entity();

void updateworld(int millis)
{
    static lastmillis = 0;
    if(lastmillis)
    {
        int time = millis - lastmillis;
        player1->input(move, strafe, time, mouseyaw, mousepitch);
    };
    lastmillis = millis;
};

void mousemove(int dx, int dy)
{
    mouseyaw += ((((double)dx)/40.0)*mousesens);
    mousepitch -= ((((double)dy)/40.0)*mousesens);
    if(mousepitch>90.0) mousepitch = 90.0;
    if(mousepitch<-90.0) mousepitch = -90.0;
    while(mouseyaw<0.0) mouseyaw += 360.0;
    while(mouseyaw>=360.0) mouseyaw -= 360.0;
};

void mouseclick(int isdown, int butnum)
{
    if(editmode && butnum==1)
    {
        editdrag(isdown!=0);
    };
};

void movep(int n) { move = n; };
void strafep(int n) { strafe = n; };

bool textures[256];
int texnums[256];
int texx[256];
int texy[256];

void purgetextures()
{
    for(int i = 0; i<256; i++) textures[i] = false;
};

#define MAXHASH 63997
#define HLOOP 5
int hash[MAXHASH+HLOOP];

void clearhash() { /*for(int i = 0; i<MAXHASH+HLOOP; i++) hash[i] = -1;*/ };

#define MAXVERTS 0x20000
#define MAXIDX 0x20000
#define VAVERTSIZE 24
float verts[MAXVERTS][5];
uchar vertc[MAXVERTS][4];
int idxs[MAXIDX];
int curvert, curidx;
float *varbuf = NULL;

inline void vert(int v1, int v2, int v3, uchar c, float t1, float t2)
{
    if(curidx==MAXIDX) fatal("too many idx");
/*
    int h = (v1+(v2<<4)+(v3<<8))%MAXHASH;
    for(int i = 0; i<HLOOP; i++, h++)
    {
        if(hash[h]<0)
        {
            hash[h] = curvert;
            break;
        }
        else
        {
            float *f = verts[hash[h]];
            if(f[2]==v1 && f[3]==v2 && f[4]==v3)
            {
                idxs[curidx++] = hash[h];
                return;
            };
        };
    };
*/
    if(curvert==MAXVERTS) fatal("too many verts");
    idxs[curidx++] = curvert;
    float *f = verts[curvert];
    f[2] = (float)v1;
    f[3] = (float)v2;
    f[4] = (float)v3;
    f[0] = t1;
    f[1] = t2;
    uchar *u = vertc[curvert++];
    u[0] = c;
    u[1] = c;
    u[2] = c;
    u[3] = 255;
};

PFNGLLOCKARRAYSEXTPROC LockArrays = NULL;
PFNGLUNLOCKARRAYSEXTPROC UnLockArrays = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC ClientActiveTexture = NULL;
PFNGLVERTEXARRAYRANGENVPROC glVAR = NULL; 
PFNWGLALLOCATEMEMORYNVPROC wglAllocateMemoryNV = NULL;
PFNWGLFREEMEMORYNVPROC wglFreeMemoryNV = NULL;

void gl_init(int w, int h)
{
    #define fogvalues 0.5f, 0.45f, 0.3f, 1.0f

    glViewport(0, 0, w, h);
    glClearColor(fogvalues);
    glClearDepth(1.0);
    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);
    
    
    glEnable(GL_FOG);
    glFogi(GL_FOG_MODE, GL_LINEAR);
    glFogf(GL_FOG_DENSITY, 0.25);
    glFogf(GL_FOG_START, 32);
    glFogf(GL_FOG_END, 256);
    glHint(GL_FOG_HINT, GL_DONT_CARE);
    GLfloat fogcolor[4] = { fogvalues };
    glFogfv(GL_FOG_COLOR, fogcolor);
    

    glEnable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_POLYGON_OFFSET_LINE);
    glPolygonOffset(-3.0, -3.0);

    //glCullFace(GL_BACK); glEnable(GL_CULL_FACE);

    char *myExtensions = (char *)glGetString(GL_EXTENSIONS);

    if(strstr(myExtensions, "GL_EXT_compiled_vertex_array"))
    {     
        LockArrays = (PFNGLLOCKARRAYSEXTPROC)wglGetProcAddress("glLockArraysEXT");
        UnLockArrays = (PFNGLUNLOCKARRAYSEXTPROC)wglGetProcAddress("glUnlockArraysEXT");
        conoutf("CVA enabled %x %x", (int)LockArrays, (int)UnLockArrays);
    };
    if(strstr(myExtensions, "GL_ARB_multitexture"))
    {     
        ClientActiveTexture = (PFNGLCLIENTACTIVETEXTUREARBPROC)wglGetProcAddress("glClientActiveTextureARB");
        conoutf("mtex enabled %x", (int)ClientActiveTexture);
    };

    if(strstr(myExtensions, "GL_NV_vertex_array_range"))
    { 
        glVAR = (PFNGLVERTEXARRAYRANGENVPROC)wglGetProcAddress("glVertexArrayRangeNV"); 
        wglAllocateMemoryNV = (PFNWGLALLOCATEMEMORYNVPROC)GLH_EXT_GET_PROC_ADDRESS("wglAllocateMemoryNV");
        wglFreeMemoryNV = (PFNWGLFREEMEMORYNVPROC)GLH_EXT_GET_PROC_ADDRESS("wglFreeMemoryNV");
        conoutf("VAR %x %x %x", (int)glVAR, (int)wglAllocateMemoryNV, (int)wglFreeMemoryNV);
        if(wglAllocateMemoryNV)
        {
            varbuf = (float *)(*wglAllocateMemoryNV)(MAXVERTS*VAVERTSIZE, 0.0f, 0.0f, 1.0f);
            if(varbuf)
            {
                glEnableClientState(GL_VERTEX_ARRAY_RANGE_NV);
                conoutf("VAR enabled %x", (int)varbuf);
            };
        };
    };   

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    //if(ClientActiveTexture) (*ClientActiveTexture)(GL_TEXTURE0_ARB);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    //if(ClientActiveTexture) (*ClientActiveTexture)(GL_TEXTURE1_ARB);
    //glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    purgetextures();
};

void cleangl()
{
    if(varbuf) (*wglFreeMemoryNV)(varbuf);
};
void installtex(int tnum, char *texname, int &xs, int &ys)
{
    SDL_Surface *s = SDL_LoadBMP(texname);
    if(!s) fatal("couldn't load texture: ", texname);
    uchar *buf = (uchar *)alloc(s->w*s->h*3);
    uchar *src = (uchar *)s->pixels;
    uchar *b = buf;
    for(int y = 0; y<s->h; y++) for(int x = 0; x<s->w; x++)
    {
        SDL_Color *c = &s->format->palette->colors[*src++];
        *b++ = c->r;
        *b++ = c->g;
        *b++ = c->b;
    };
    //glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,tnum);
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); 
    if(gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, s->w, s->h, GL_RGB, GL_UNSIGNED_BYTE, buf)) fatal("could not build mipmaps");
    //glDisable(GL_TEXTURE_2D);
    xs = s->w;
    ys = s->h;
    SDL_FreeSurface(s);
};

char *skynames[] =
{
    "iklake_bk",
    "iklake_dn",
    "iklake_ft",
    "iklake_lf",
    "iklake_rt",
    "iklake_up",
    NULL
};

char *texturenames[] =
{
    NULL,                       // 0: transparent/sky
    "ik_brick_3216d",           // 1: default wall
    //"stone1",
    "ik_floor_brick128a",       // 2: default floor
    "ik_wood_planks128a",       // 3: default ceil
    "ik_brickwall_red",
    "ik_brick_12832a",          
    "ik_brick_12864a",
    "ik_brick_128128a",
    "ik_brick_trimh16b",
    "ik_floor_brick128b",
    "ik_floor_tile64b",
    "ik_floor_wood128d",
    "ik_ground_cobbles128",
    "ik_ground_grass128",
    "ik_ground_lawn128",
    "ik_ground_pavement128",
    "ik_tree128moss",
    "ik_trim_brickh32a", //17
    NULL
}; 

int lookuptexture(int tex)
{
    if(textures[tex]) return texnums[tex];
    if(tex>17 || tex==0) return 1;
    char name[256];
    sprintf(name, "data%ciktex%c%s.bmp", PATHDIV, PATHDIV, texturenames[tex]);
    int tnum = tex+10, xs, ys;
    installtex(tnum, name, xs, ys);
    textures[tex] = true;
    texnums[tex] = tnum;
    texx[tex] = xs;
    texy[tex] = ys;
    return tnum;
};

#define MAXSTRIPS 0x10000
int stript[MAXSTRIPS], strips[MAXSTRIPS], stripn[MAXSTRIPS], nstrips;

void wastetime()
{
    float sum =  0;
    for(int i = 0; i<MAXVERTS; i++) for(int j = 0; j<5; j++) sum += verts[i][j]*5/11;
};

void renderstrips()
{
    if(varbuf)
    {
        //wastetime();
        // sync happens here?
        (*glVAR)(MAXVERTS*VAVERTSIZE, varbuf);
        float *p = varbuf;
        for(int i = 0; i<curvert; i++)
        {
            for(int j = 0; j<5; j++) *p++ = verts[i][j];
            int *pi = (int *)p++;
            *pi = *((int *)(&vertc[i][0])); 
        };
        glVertexPointer(3, GL_FLOAT, VAVERTSIZE, varbuf+2);
        glColorPointer(4, GL_UNSIGNED_BYTE, VAVERTSIZE, varbuf+5);
        glTexCoordPointer(2, GL_FLOAT, VAVERTSIZE, varbuf);
        //uchar isvalid;
        //glGetBooleanv(GL_VERTEX_ARRAY_RANGE_VALID_NV, &isvalid);
        //conoutf("valid = %d", isvalid);
    }
    else
    {
        glVertexPointer(3, GL_FLOAT, 20, &verts[0][2]);
        glColorPointer(4, GL_UNSIGNED_BYTE, 4, &vertc);
        //if(ClientActiveTexture) (*ClientActiveTexture)(GL_TEXTURE0_ARB);
        glTexCoordPointer(2, GL_FLOAT, 20, verts);
        //if(ClientActiveTexture) (*ClientActiveTexture)(GL_TEXTURE1_ARB);
        //glTexCoordPointer(2, GL_FLOAT, 28, &verts[0][2]);
        //if(LockArrays) (*LockArrays)(0, curvert);
    };

    for(int j = 0; j<nstrips; j++)
    {
        glBindTexture(GL_TEXTURE_2D, lookuptexture(stript[j]));
        glDrawElements(GL_QUAD_STRIP, stripn[j], GL_UNSIGNED_INT, &idxs[strips[j]]);
        //glDrawArrays(GL_QUAD_STRIP, strips[j], stripn[j]);
    };

    //if(UnLockArrays) (*UnLockArrays)();
/*    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
*/
};

void strip(int tex, int start, int n)
{
    if(nstrips==MAXSTRIPS) fatal("too many strips");
    stript[nstrips] = tex;
    strips[nstrips] = start;
    stripn[nstrips++] = n;
};

int nquads;
#define TEXTURESCALE 32.0f
int floorstrip = false;
int otex, oh, oy, oskip;
int ol3, ol4;
int firstindex;
#define LIGHT_EPSILON 10  // tune these 2 for optimal light quality vs speed
#define LIGHT_SKIP 10
inline int fabs(int x) { return x<0 ? -x : x; };

void stripend()
{
    if(floorstrip)
    {
        strip(otex, firstindex, curidx-firstindex);
        floorstrip = false;
    };
};

void render_flat(int wtex, int x, int y, int size, int h, uchar l1, uchar l2, uchar l3, uchar l4)
{
    float xf = TEXTURESCALE/texx[wtex];
    float yf = TEXTURESCALE/texy[wtex];
    float xs = size*xf;
    float ys = size*yf;
    float xo = xf*x;
    float yo = yf*y;

    bool first = !floorstrip || y!=oy+size || wtex!=otex || h!=oh;

    if(first)
    {
        stripend();
        firstindex = curidx;
        otex = wtex;
        oh = h;
        floorstrip = true;
        vert(x,      h, y, l1, xo,    yo);
        vert(x+size, h, y, l2, xo+xs, yo);
        ol3 = l1;
        ol4 = l2;
        oskip = 0;
    }
    else
    {
        if(!fisheye && oskip<LIGHT_SKIP && fabs(ol3-l3)<LIGHT_EPSILON && fabs(ol4-l4)<LIGHT_EPSILON)
        {
            curidx -= 2;
            if(idxs[curidx+1]==curvert-1) curvert--;
            if(idxs[curidx]==curvert-1) curvert--;
            // and remove from hash!
            // avoid generating them in the first place?
            nquads--;
            oskip++;
        }
        else
        {
            oskip = 0;
        };
    };

    vert(x,      h, y+size, l4, xo,    yo+ys);
    vert(x+size, h, y+size, l3, xo+xs, yo+ys); 

    oy = y;
    nquads++;
};

void render_square(int wtex, int floor, int ceil, int x1, int y1, int x2, int y2, int size, uchar l1, uchar l2)
{
    stripend();

    if(!fisheye)
    {
        float xf = TEXTURESCALE/texx[wtex];
        float yf = TEXTURESCALE/texy[wtex];
        float xs = size*xf;
        float ys = (ceil-floor)*yf;
        float xo = xf*(x1==x2 ? min(y1,y2) : min(x1,x2));
        float yo = yf*ceil;
        vert(x1, ceil,  y1, l1, xo,    yo);
        vert(x2, ceil,  y2, l2, xo+xs, yo); 
        vert(x1, floor, y1, l1, xo,    yo+ys); 
        vert(x2, floor, y2, l2, xo+xs, yo+ys); 
        nquads++;
        strip(wtex, curidx-4, 4);
    }
    else
    {
        glBindTexture(GL_TEXTURE_2D, lookuptexture(wtex));
        glBegin(GL_QUAD_STRIP);
        float xf = TEXTURESCALE/texx[wtex];
        float yf = TEXTURESCALE/texy[wtex];
        float xs = size*xf;
        float ys = size*yf;
        float xo = xf*(x1==x2 ? min(y1,y2) : min(x1,x2));

        for(int f = floor; f<=ceil; f += size)
        {
            float yo = yf*f; //convert to vertex arrays!
            glColor3f(l1,l1,l1); glTexCoord2f(xo, yo);       glVertex3i(x1, f, y1);
            glColor3f(l2,l2,l2); glTexCoord2f(xo+xs, yo);    glVertex3i(x2, f, y2);
        };

        nquads += (ceil-floor)/size;
        glEnd();
    };
};

void line(int x1, int y1, int z1, int x2, int y2, int z2, float width, bool stipple, int c)
{
    glLineWidth(width);
    /*
    if(stipple)
    {
        glLineStipple(1, 255);
        glEnable(GL_LINE_STIPPLE);
    }
    else
    {
        glDisable(GL_LINE_STIPPLE);
    };
    */
    glColor3ub(c,c,c);
    glColor3f(1.0f,1.0f,1.0f);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glBegin(GL_POLYGON);
    glVertex3i(x1, z1, y1);
    glVertex3i(x1+1, z1, y1);
    glVertex3i(x1+1, z1+1, y1);
    glVertex3i(x1, z1+1, y1);
    //glVertex3i(x2, z2, y2);
    glEnd();
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
};

void box(int x, int y, int z, int w, int h, float width, bool stipple, int c)
{
    glLineWidth(width);
    /*
    if(stipple)
    {
        glLineStipple(1, 255);
        glEnable(GL_LINE_STIPPLE);
    }
    else
    {
        glDisable(GL_LINE_STIPPLE);
    };
    */
    glColor3ub(c,c,c);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glBegin(GL_POLYGON);
    glVertex3i(x, z, y);
    glVertex3i(x+w, z, y);
    glVertex3i(x+w, z, y+h);
    glVertex3i(x, z, y+h);
    glEnd();
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
};

void obtaincursor(int x, int y)
{
    float depth;
    glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &depth);
    GLint viewport[4];
    GLdouble mm[16], pm[16];
    glGetIntegerv(GL_VIEWPORT, viewport);
    glGetDoublev(GL_MODELVIEW_MATRIX, mm);
    glGetDoublev(GL_PROJECTION_MATRIX, pm);
    double wx, wy, wz;
    gluUnProject(x, y, depth, mm, pm, viewport, &wx, &wz, &wy);
    cursorupdate(wx, wy, wz);
};

void edittoggle() { editmode = !editmode; };

void gl_drawhud(int w, int h, int curfps)
{
	glDisable(GL_DEPTH_TEST);

    if(editmode) obtaincursor(w/2, h/2);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glOrtho(0, 800, 600, 0, -1, 1);

    draw_text("fps %d", 650, 10, 2, curfps, 0);
    draw_text("LOD %d", 650, 35, 2, lod_factor(), 0);
    draw_text("qds %d", 650, 60, 2, nquads, 0);
    draw_text("vts %d", 650, 85, 2, curvert, 0);
    //draw_text("(x,y) = (%d,%d)", 700, 70, 2, wx, wy);

    renderconsole();

	glLoadIdentity();
	glOrtho(0, w*2, h*2, 0, -1, 1);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_SRC_ALPHA);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, 1);
	glBegin(GL_QUADS);
	glTexCoord2d(0.0, 1.0); glVertex2d(w - 8.0, h + 8.0);
	glTexCoord2d(1.0, 1.0); glVertex2d(w + 8.0, h + 8.0);
	glTexCoord2d(1.0, 0.0); glVertex2d(w + 8.0, h - 8.0);
	glTexCoord2d(0.0, 0.0); glVertex2d(w - 8.0, h - 8.0);
	glEnd();
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);

	glEnable(GL_DEPTH_TEST);
};

double rad(double x) { return x*3.14159/180.0; };

void gl_drawframe(int w, int h, int changelod, float curfps)
{
    nquads = 0;

    camx += move*cos(rad(mouseyaw-90.0))*cos(rad(mousepitch));
    camz += move*sin(rad(mouseyaw-90.0))*cos(rad(mousepitch));
    camx += strafe*cos(rad(mouseyaw-180.0));
    camz += strafe*sin(rad(mouseyaw-180.0));

    camy += move*sin(rad(mousepitch));

    //glNewList(1, GL_COMPILE);

    glMatrixMode(GL_PROJECTION);
    gluPerspective(80.0f,(GLfloat)w/(GLfloat)h,0.1f,5000.0f);

    glClear(/*GL_COLOR_BUFFER_BIT |*/ GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glRotated(mousepitch,-1.0,0.0,0.0);
    glRotated(mouseyaw,0.0,1.0,0.0);
    glTranslated(-camx,-camy,-camz);	

    glEnable(GL_TEXTURE_2D);
    floorstrip = false;
    curvert = 0;
    curidx = 0;
    nstrips = 0;
    clearhash();
    render_world((int)camx, (int)camz, (int)camy, changelod, (int)mouseyaw);
    stripend();
    renderstrips();
    glDisable(GL_TEXTURE_2D);

    //glEndList();
    //glCallList(1);


    glColor3ub(255,255,255);

    gl_drawhud(w,h,(int)curfps);
};

