// the engine

#include "cube.h"

#ifdef WIN32
#include "windows.h"
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>
#include <glh/glh_extensions.h>

#include <stdio.h>
#include <stdlib.h>

#include <SDL.h>

bool fisheye = false;
bool useVAR = true;

void pre_gfx_init()
{
    empty_world();
    player1->x = player1->y = (float)size/2;
};

bool editmode = false; 

bool textures[256];
int texnums[256];
int texx[256];
int texy[256];

void purgetextures()
{
    for(int i = 0; i<256; i++) textures[i] = false;
};

#define MAXVERTS 0x20000
#define VAVERTSIZE (sizeof(float)*6)
float verts[MAXVERTS][6];
int curvert;
float *varbuf = NULL;

inline void vert(int v1, int v2, int v3, uchar c, float t1, float t2)
{
    if(curvert==MAXVERTS) fatal("too many verts");
    float *f = verts[curvert++];
    *f++ = t1;
    *f++ = t2;
    *f++ = (float)v1;
    *f++ = (float)v2;
    *f++ = (float)v3;
    int *i = (int *)f;
    *i = c | c<<8 | c<<16 | 0xFF000000;
};

PFNGLLOCKARRAYSEXTPROC LockArrays = NULL;
PFNGLUNLOCKARRAYSEXTPROC UnLockArrays = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC ClientActiveTexture = NULL;
PFNGLVERTEXARRAYRANGENVPROC glVAR = NULL; 
PFNWGLALLOCATEMEMORYNVPROC wglAllocateMemoryNV = NULL;
PFNWGLFREEMEMORYNVPROC wglFreeMemoryNV = NULL;

void gl_init(int w, int h)
{
    #define fogvalues 0.5f, 0.45f, 0.3f, 1.0f

    glViewport(0, 0, w, h);
    glClearColor(fogvalues);
    glClearDepth(1.0);
    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);
    
    
    glEnable(GL_FOG);
    glFogi(GL_FOG_MODE, GL_LINEAR);
    glFogf(GL_FOG_DENSITY, 0.25);
    glFogf(GL_FOG_START, 32);
    glFogf(GL_FOG_END, 256);
    glHint(GL_FOG_HINT, GL_DONT_CARE);
    GLfloat fogcolor[4] = { fogvalues };
    glFogfv(GL_FOG_COLOR, fogcolor);
    

    glEnable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_POLYGON_OFFSET_LINE);
    glPolygonOffset(-3.0, -3.0);

    //glCullFace(GL_BACK); glEnable(GL_CULL_FACE);

    char *myExtensions = (char *)glGetString(GL_EXTENSIONS);

    if(useVAR && strstr(myExtensions, "GL_NV_vertex_array_range"))
    { 
        glVAR = (PFNGLVERTEXARRAYRANGENVPROC)wglGetProcAddress("glVertexArrayRangeNV"); 
        wglAllocateMemoryNV = (PFNWGLALLOCATEMEMORYNVPROC)GLH_EXT_GET_PROC_ADDRESS("wglAllocateMemoryNV");
        wglFreeMemoryNV = (PFNWGLFREEMEMORYNVPROC)GLH_EXT_GET_PROC_ADDRESS("wglFreeMemoryNV");

        if(wglAllocateMemoryNV)
        {
            char *vartype = "AGP memory";
            varbuf = (float *)(*wglAllocateMemoryNV)(MAXVERTS*VAVERTSIZE, 0.0f, 0.0f, 0.5f);

            if(!varbuf)
            {
                varbuf = (float *)(*wglAllocateMemoryNV)(MAXVERTS*VAVERTSIZE, 0.0f, 0.0f, 1.0f);
                vartype = "on-card video memory";
            };

            if(varbuf)
            {
                glEnableClientState(GL_VERTEX_ARRAY_RANGE_NV);
                conoutf("vertex_array_range set up (using %s)", (int)vartype);
            };
        };
    };   

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    purgetextures();
};

void cleangl()
{
    if(varbuf) (*wglFreeMemoryNV)(varbuf);
};
void installtex(int tnum, char *texname, int &xs, int &ys)
{
    SDL_Surface *s = SDL_LoadBMP(texname);
    if(!s) fatal("couldn't load texture: ", texname);
    uchar *buf = (uchar *)alloc(s->w*s->h*3);
    uchar *src = (uchar *)s->pixels;
    uchar *b = buf;
    for(int y = 0; y<s->h; y++) for(int x = 0; x<s->w; x++)
    {
        SDL_Color *c = &s->format->palette->colors[*src++];
        *b++ = c->r;
        *b++ = c->g;
        *b++ = c->b;
    };
    glBindTexture(GL_TEXTURE_2D,tnum);
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); 
    if(gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, s->w, s->h, GL_RGB, GL_UNSIGNED_BYTE, buf)) fatal("could not build mipmaps");
    xs = s->w;
    ys = s->h;
    SDL_FreeSurface(s);
};

char *skynames[] =
{
    "iklake_bk",
    "iklake_dn",
    "iklake_ft",
    "iklake_lf",
    "iklake_rt",
    "iklake_up",
    NULL
};

char *texturenames[] =
{
    NULL,                       // 0: transparent/sky
    "ik_brick_3216d",           // 1: default wall
    //"stone1",
    "ik_floor_brick128a",       // 2: default floor
    "ik_wood_planks128a",       // 3: default ceil
    "ik_brickwall_red",
    "ik_brick_12832a",          
    "ik_brick_12864a",
    "ik_brick_128128a",
    "ik_brick_trimh16b",
    "ik_floor_brick128b",
    "ik_floor_tile64b",
    "ik_floor_wood128d",
    "ik_ground_cobbles128",
    "ik_ground_grass128",
    "ik_ground_lawn128",
    "ik_ground_pavement128",
    "ik_tree128moss",
    "ik_trim_brickh32a", //17
    NULL
}; 

int lookuptexture(int tex)
{
    if(textures[tex]) return texnums[tex];
    if(tex>17 || tex==0) return 1;
    char name[256];
    sprintf(name, "data%ciktex%c%s.bmp", PATHDIV, PATHDIV, texturenames[tex]);
    int tnum = tex+10, xs, ys;
    installtex(tnum, name, xs, ys);
    textures[tex] = true;
    texnums[tex] = tnum;
    texx[tex] = xs;
    texy[tex] = ys;
    return tnum;
};

#define MAXSTRIPS 0x10000
int stript[MAXSTRIPS], strips[MAXSTRIPS], stripn[MAXSTRIPS], nstrips;

void renderstrips()
{
    float *buf = &verts[0][0];

    if(varbuf)
    {
        (*glVAR)(MAXVERTS*VAVERTSIZE, varbuf); // sync happens here too
        memcpy(varbuf, verts, curvert*VAVERTSIZE);
        buf = varbuf;
    };

    glVertexPointer(3, GL_FLOAT, VAVERTSIZE, buf+2);
    glColorPointer(4, GL_UNSIGNED_BYTE, VAVERTSIZE, buf+5);
    glTexCoordPointer(2, GL_FLOAT, VAVERTSIZE, buf);

    for(int j = 0; j<nstrips; j++)
    {
        glBindTexture(GL_TEXTURE_2D, lookuptexture(stript[j]));
        glDrawArrays(GL_QUAD_STRIP, strips[j], stripn[j]);
    };
};

void strip(int tex, int start, int n)
{
    if(nstrips==MAXSTRIPS) fatal("too many strips");
    stript[nstrips] = tex;
    strips[nstrips] = start;
    stripn[nstrips++] = n;
};

int nquads;
#define TEXTURESCALE 32.0f
int floorstrip = false;
int otex, oh, oy, oskip;
int ol3, ol4;
int firstindex;
#define LIGHT_EPSILON 10  // tune these 2 for optimal light quality vs speed
#define LIGHT_SKIP 10
inline int fabs(int x) { return x<0 ? -x : x; };

void stripend()
{
    if(floorstrip)
    {
        strip(otex, firstindex, curvert-firstindex);
        floorstrip = false;
    };
};

void render_flat(int wtex, int x, int y, int size, int h, uchar l1, uchar l2, uchar l3, uchar l4)
{
    float xf = TEXTURESCALE/texx[wtex];
    float yf = TEXTURESCALE/texy[wtex];
    float xs = size*xf;
    float ys = size*yf;
    float xo = xf*x;
    float yo = yf*y;

    bool first = !floorstrip || y!=oy+size || wtex!=otex || h!=oh;

    if(first)
    {
        stripend();
        firstindex = curvert;
        otex = wtex;
        oh = h;
        floorstrip = true;
        vert(x,      h, y, l1, xo,    yo);
        vert(x+size, h, y, l2, xo+xs, yo);
        ol3 = l1;
        ol4 = l2;
        oskip = 0;
    }
    else
    {
        if(!fisheye && oskip<LIGHT_SKIP && fabs(ol3-l3)<LIGHT_EPSILON && fabs(ol4-l4)<LIGHT_EPSILON)
        {
            curvert -= 2;
            // avoid generating them in the first place?
            nquads--;
            oskip++;
        }
        else
        {
            oskip = 0;
        };
    };

    vert(x,      h, y+size, l4, xo,    yo+ys);
    vert(x+size, h, y+size, l3, xo+xs, yo+ys); 

    oy = y;
    nquads++;
};

void render_square(int wtex, int floor, int ceil, int x1, int y1, int x2, int y2, int size, uchar l1, uchar l2)
{
    stripend();

    if(!fisheye)
    {
        float xf = TEXTURESCALE/texx[wtex];
        float yf = TEXTURESCALE/texy[wtex];
        float xs = size*xf;
        float ys = (ceil-floor)*yf;
        float xo = xf*(x1==x2 ? min(y1,y2) : min(x1,x2));
        float yo = yf*ceil;
        vert(x1, ceil,  y1, l1, xo,    yo);
        vert(x2, ceil,  y2, l2, xo+xs, yo); 
        vert(x1, floor, y1, l1, xo,    yo+ys); 
        vert(x2, floor, y2, l2, xo+xs, yo+ys); 
        nquads++;
        strip(wtex, curvert-4, 4);
    }
    else
    {
        glBindTexture(GL_TEXTURE_2D, lookuptexture(wtex));
        glBegin(GL_QUAD_STRIP);
        float xf = TEXTURESCALE/texx[wtex];
        float yf = TEXTURESCALE/texy[wtex];
        float xs = size*xf;
        float ys = size*yf;
        float xo = xf*(x1==x2 ? min(y1,y2) : min(x1,x2));

        for(int f = floor; f<=ceil; f += size)
        {
            float yo = yf*f; //convert to vertex arrays!
            glColor3f(l1,l1,l1); glTexCoord2f(xo, yo);       glVertex3i(x1, f, y1);
            glColor3f(l2,l2,l2); glTexCoord2f(xo+xs, yo);    glVertex3i(x2, f, y2);
        };

        nquads += (ceil-floor)/size;
        glEnd();
    };
};

void line(int x1, int y1, int z1, int x2, int y2, int z2)
{
    glBegin(GL_POLYGON);
    glVertex3i(x1, z1, y1);
    glVertex3f((float)x1, (float)z1, y1+0.01f);
    glVertex3f((float)x2, (float)z2, y2+0.01f);
    glVertex3i(x2, z2, y2);
    glEnd();
};

void linestyle(float width, int r, int g, int b)
{
    glLineWidth(width);
    glColor3ub(r,g,b);
};

void box(int x, int y, int z, int w, int h)
{
    glBegin(GL_POLYGON);
    glVertex3i(x, z, y);
    glVertex3i(x+w, z, y);
    glVertex3i(x+w, z, y+h);
    glVertex3i(x, z, y+h);
    glEnd();
};

void obtaincursor(int x, int y)
{
    float depth;
    glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &depth);
    GLint viewport[4];
    GLdouble mm[16], pm[16];
    glGetIntegerv(GL_VIEWPORT, viewport);
    glGetDoublev(GL_MODELVIEW_MATRIX, mm);
    glGetDoublev(GL_PROJECTION_MATRIX, pm);
    double wx, wy, wz;
    gluUnProject(x, y, depth, mm, pm, viewport, &wx, &wz, &wy);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    cursorupdate(wx, wy, wz);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
};

void gl_drawhud(int w, int h, int curfps)
{
    if(editmode) obtaincursor(w/2, h/2);

	glDisable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glOrtho(0, 800, 600, 0, -1, 1);

    glColor3ub(255,255,255);

    draw_text("fps %d", 650, 10, 2, curfps, 0);
    draw_text("LOD %d", 650, 35, 2, lod_factor(), 0);
    draw_text("qds %d", 650, 60, 2, nquads, 0);
    draw_text("vts %d", 650, 85, 2, curvert, 0);
    //draw_text("(x,y) = (%d,%d)", 700, 70, 2, wx, wy);

    renderconsole();

	glLoadIdentity();
	glOrtho(0, w*2, h*2, 0, -1, 1);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_SRC_ALPHA);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, 1);
	glBegin(GL_QUADS);
	glTexCoord2d(0.0, 1.0); glVertex2d(w - 8.0, h + 8.0);
	glTexCoord2d(1.0, 1.0); glVertex2d(w + 8.0, h + 8.0);
	glTexCoord2d(1.0, 0.0); glVertex2d(w + 8.0, h - 8.0);
	glTexCoord2d(0.0, 0.0); glVertex2d(w - 8.0, h - 8.0);
	glEnd();
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);

	glEnable(GL_DEPTH_TEST);
};

void gl_drawframe(int w, int h, int changelod, float curfps)
{
    if(player1->outsidemap)
    {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }
    else
    {
        glClear(GL_DEPTH_BUFFER_BIT);
    };

    glMatrixMode(GL_PROJECTION);
    gluPerspective(80.0f,(GLfloat)w/(GLfloat)h,0.1f,5000.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glRotated(player1->pitch,-1.0,0.0,0.0);
    glRotated(player1->yaw,0.0,1.0,0.0);
    glTranslated(-player1->x,-player1->z,-player1->y);	

    glEnable(GL_TEXTURE_2D);
    floorstrip = false;
    nquads = 0;
    curvert = 0;
    nstrips = 0;
    render_world((int)player1->x, (int)player1->y, (int)player1->z, changelod, (int)player1->yaw);
    stripend();
    renderstrips();
    glDisable(GL_TEXTURE_2D);

    gl_drawhud(w,h,(int)curfps);
};

