#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "cube.h"

sqr *world;
int sfactor, size, csize;

header *hdr;

int rnd(int max) { return rand()%max; };

void remip(int level, int x, int y, int xs, int ys)
{
    if(level>=SMALLEST_FACTOR) return;
    sqr *w = MIP(level);
    sqr *v = MIP(level+1);
    int ws = size>>level;
    int vs = size>>(level+1);
    if(x&1) { x--; xs++; };
    if(y&1) { y--; ys++; };
    xs = (xs+1)&~1;
    ys = (ys+1)&~1;
    for(int xx = x; xx<x+xs; xx+=2) for(int yy = y; yy<y+ys; yy+=2)
    {
        sqr *o[4];
        o[0] = SWS(w,xx,yy,ws);
        o[1] = SWS(w,xx+1,yy,ws);
        o[2] = SWS(w,xx+1,yy+1,ws);
        o[3] = SWS(w,xx,yy+1,ws);
        sqr *r = SWS(v,xx/2,yy/2,vs);
        r->type = SOLID(o[0])+SOLID(o[1])+SOLID(o[2])+SOLID(o[3])>1 ? SOLID : SPACE;
        r->wtex = o[0]->wtex; // see if a texture occurs >1, or solid ?
        if(!SOLID(r))
        {
            int floor = -128, ceil = 127, ftex, ctex, light = 0, num = 0;
            for(int i = 0; i<4; i++) if(!SOLID(o[i]))
            {
                num++;
                light += o[i]->light;
                if(o[i]->floor>floor) floor = o[i]->floor;
                if(o[i]->ceil<ceil) ceil = o[i]->ceil;
                ftex = o[i]->ftex;
                ctex = o[i]->ctex;
            };
            int bit = 1<<level;
            r->floor = floor&bit ? floor-bit : floor;
            r->ceil = ceil&bit ? ceil+bit : ceil;
            r->ftex = ftex;
            r->ctex = ctex;
            r->light = light/num;
        };
    };
    remip(level+1, x/2, y/2, xs/2, ys/2);
};

void excavate(int x, int y, int xs, int ys, int floor, int ceil)
{
    if(x<MINBORD) x = MINBORD;
    if(y<MINBORD) y = MINBORD;
    if(x+xs>size-MINBORD) xs = size-MINBORD-x;
    if(y+ys>size-MINBORD) ys = size-MINBORD-y;
    for(int xa = 0; xa<xs; xa++) for(int ya = 0; ya<ys; ya++)
    {
        sqr *s = S(x+xa,y+ya);
        s->type = SPACE;
        s->floor = floor;
        s->ceil = ceil;
        s->ftex = 2;
        s->ctex = 3;
        s->wtex = 1;
        s->light = 128;
    }; 
    remip(0, x, y, xs, ys);
};

void lightray(int lx, int ly, int bx, int by, int reach)
{
    int dx = bx-lx;
    int dy = by-ly;
    double dist = sqrt(dx*dx+dy*dy);
    if(dist>=reach) return;
    int steps = reach*4;
    double stepx = dx/(double)steps;
    double stepy = dy/(double)steps;
    double x = lx;
    double y = ly;
    for(int i = 0; i<steps; i++)
    {
        if(SOLID(S((int)x,(int)y))) return;
        x += stepx;
        y += stepy;
    };
    sqr *s = S(bx,by);
    int ll = 255-(int)((dist/(double)reach)*255.0);
    if(ll>s->light) s->light = ll;
};

void calclight()
{
    for(int x = 0; x<size; x++) for(int y = 0; y<size; y++) S(x,y)->light = 25;
    for(int i = 0; i<MAXENTS; i++)
    {
        entity *e = &hdr->ents[i];
        if(e->type==LIGHT)
        {
            int reach = e->attr;
            int sx = e->x-reach;
            int ex = e->x+reach;
            int sy = e->y-reach;
            int ey = e->y+reach;
            if(sx<1) sx = 1;
            if(ex<1) ex = 1;
            if(sy<1) sy = 1;
            if(ey<1) ey = 1;
            if(sx>size-2) sx = size-2;
            if(ex>size-2) ex = size-2;
            if(sy>size-2) sy = size-2;
            if(ey>size-2) ey = size-2;
            for(;sx<=ex; sx++) for(int sy2 = sy ;sy2<=ey; sy2++)
            {
                lightray(e->x, e->y, sx, sy2, reach);
            };
        };
    };
    remip(0, 0, 0, size, size);
};

void clearents(int type)
{
    for(int i = 0; i<MAXENTS; i++)
    {
        entity *e = &hdr->ents[i];
        if(e->type==type) e->type = NOTUSED;
    };
};

entity *newentity(int x, int y, int z, int type)
{
    for(int i = 0; i<MAXENTS; i++)
    {
        entity *e = &hdr->ents[i];
        if(e->type==NOTUSED)
        {
            e->type = type;
            e->x = x;
            e->y = y;
            e->z = z;

            switch(type)
            {
                case LIGHT:
                    e->attr = 16;
                    break;
            };
            return e;
        };
    };
    return hdr->ents;
};

void empty_world()
{
    sfactor = 8;
    size = 1<<sfactor;
    csize = size*size;
    world = (sqr *)alloc(csize*sizeof(sqr)*2);
    for(int x = 0; x<size; x++) for(int y = 0; y<size; y++)
    {
        sqr *s = S(x,y);
        s->light = 25;
        s->wtex = 1;//4+((x*y*11+7)&7); //1;
        s->ftex = 2;
        s->ctex = 3;
        s->type = SOLID;
        s->floor = 0;
        s->ceil = 8;
    };
    hdr = (header *)alloc(sizeof(header));
    strncpy(hdr->head, "CUBE", 4);
    strncpy(hdr->tail, "EBUC", 4);
    hdr->version = MAPVERSION;
    hdr->headersize = sizeof(header);
    strcpy(hdr->maptitle, "Untitled Map by Unknown");
    for(int t = 0; t<3; t++) for(int i = 0; i<256; i++) hdr->texlists[t][i] = i;
    int mid = size/2;
    //excavate(2,2,size-4,size-4,0,32);
    for(int i = 0; i<1000; i++) excavate(rnd(size/2)*2, rnd(size/2)*2, rnd(8)*2+2, rnd(8)*2+2, rnd(2)*2, rnd(3)*2+8);
    for(int l = 0; l<200; l++) newentity(rnd(size), rnd(size), 0, LIGHT);
    calclight();
};

