#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cube.h"

void render_wall(sqr *o, sqr *s, int x1, int y1, int x2, int y2, int mip, int l1, int l2)
{
    if(SOLID(o))
    {
        render_square(o->wtex, s->floor, s->ceil, x1<<mip, y1<<mip, x2<<mip, y2<<mip, 1<<mip, l1, l2);
    }
    else
    {
        if(o->floor>s->floor) render_square(o->wtex, s->floor, o->floor, x1<<mip, y1<<mip, x2<<mip, y2<<mip, 1<<mip, l1, l2);
        if(o->ceil<s->ceil)   render_square(o->wtex, o->ceil,  s->ceil,  x1<<mip, y1<<mip, x2<<mip, y2<<mip, 1<<mip, l1, l2);
    };
};

#define MAX_MIP SMALLEST_FACTOR
#define MIN_LOD 2
#define LOW_LOD 2
#define MAX_LOD 64 

int lod = 8, lodtop, lodbot, lodleft, lodright, lock = 0;

int lod_factor() { return lod; };
void locklod(int f) { lock = f; };

void render_seg(int vx, int vy, int vh, int mip, int x, int y, int xs, int ys)
{
    sqr *w = MIP(mip);
    int vxx = (vx+(1<<mip)/2)>>mip;
    int vyy = (vy+(1<<mip)/2)>>mip;
    int lx = vxx-lodleft;
    int ly = vyy-lodtop;
    int rx = vxx+lodright;
    int ry = vyy+lodbot;
    int sz = size>>mip;

    if(mip) render_seg(vx, vy, vh, mip-1, lx*2, ly*2, (rx-lx)*2, (ry-ly)*2);
    
    xs += x;
    ys += y;
    if(xs<0 || ys<0 || x>=sz || y>=sz) return;
    if(x<0) x = 0;
    if(y<0) y = 0;
    if(xs>=sz) xs = sz;
    if(ys>=sz) ys = sz; 

    #define PRE {for(int xx = x; xx<xs; xx++) for(int yy = y; yy<ys; yy++) \
                {if(xx>=lx && xx<rx && yy>=ly && yy<ry && mip) continue; \
                sqr *s = SWS(w,xx,yy,sz); \
                if(SOLID(s)) continue;
    #define POST };};
    #define l(x,y) (SWS(s,x,y,sz)->light)

    PRE
    if(s->floor<=vh) render_flat(s->ftex, xx<<mip, yy<<mip, 1<<mip, s->floor, s->light, l(1,0), l(1,1), l(0,1));
    POST  
      
    PRE
    if(s->ceil>=vh) render_flat(s->ctex, xx<<mip, yy<<mip, 1<<mip, s->ceil, s->light, l(1,0), l(1,1), l(0,1));
    POST

    PRE
    sqr *t = SWS(s,1,0,sz);
    sqr *u = SWS(s,1,1,sz);
    sqr *v = SWS(s,0,1,sz);
    if(xx<=vxx) render_wall(SWS(s,-1,0,sz), s, xx, yy, xx, yy+1, mip, s->light, v->light);
    if(xx>=vxx) render_wall(t,              s, xx+1, yy, xx+1, yy+1, mip, t->light, u->light);
    if(yy<=vyy) render_wall(SWS(s,0,-1,sz), s, xx, yy, xx+1, yy, mip, s->light, t->light);
    if(yy>=vyy) render_wall(v,              s, xx, yy+1, xx+1, yy+1, mip, v->light, u->light);
    POST
};

void distlod(int &low, int &high, int angle)
{
    float f = (90.0f/(float)lod)+0.1f;
    low = (int)((90-angle)/f);
    high = (int)(angle/f);
    if(low<MIN_LOD) low = MIN_LOD;
    if(high<MIN_LOD) high = MIN_LOD;
};

void render_world(int vx, int vy, int vh, int changelod, int yaw)
{
    yaw = 360-yaw;
    lod += changelod;
    if(lock) lod = lock;
    if(lod<LOW_LOD) lod = LOW_LOD;
    if(lod>MAX_LOD) lod = MAX_LOD;
    lodtop = lodbot = lodleft = lodright = MIN_LOD;
    if(yaw>45 && yaw<=135)
    {
        lodleft = lod;
        distlod(lodtop, lodbot, yaw-45);
    }
    else if(yaw>135 && yaw<=225)
    {
        lodbot = lod;
        distlod(lodleft, lodright, yaw-135);
    }
    else if(yaw>225 && yaw<=315)
    {
        lodright = lod;
        distlod(lodbot, lodtop, yaw-225);
    }
    else
    {
        lodtop = lod;
        distlod(lodright, lodleft, yaw<=45 ? yaw+45 : yaw-315);
    };
    render_seg(vx, vy, vh, MAX_MIP, 0, 0, size>>MAX_MIP, size>>MAX_MIP);
};

